#!/usr/bin/env python

# exit codes
NOT_OK = 1
OK = 0

# modules for OPUS internal poller
import sys                  # access to command-line args
import string               # string functions
import opusutil             # utilites like PrintMsg
import re                   # regular expression handling
import traceback            # printing of stack traces 

# python modules for this application
import os
import getopt
import time
import glob
import pyfits

# local modules for this application
import base_poller

# exceptions
class UsageError(Exception):
  pass

class FailedResourceInit(Exception):
  pass

class PollingFailed(Exception):
  pass

class HaltEvent(Exception):
  pass

class DatasetNotFoundInOSF(Exception):
  pass

class NoFilesFound(Exception):
  pass

class ArchivePrep(base_poller.InternalPoller):
  """
########################################################################
Class: ArchivePrep

Description:
------------
The class prepares a CDBS reference file delivery for archiving by
creating OSFs for each reference file dataset.

Members:
--------
trlpath = directory for trailer file to be updated
inpath = directory where reference files are found
archive_dir = directory where ref files will be archived from
pathname = name of path this task is running in (for OSF creation)
osf_status = values of OSF columns for archiving OSFs (e.g. ccw )

Methods:
--------
usage() - prints an interactive usage message and raises an exception

########################################################################
  """
  def __init__(self):
    # initialize members
    self._trlpath = None
    self._inpath = None
    self._pathname = None
    self._archive_dir = None
    self._osf_status = None

  def initialize(self):
    #
    # call superclass constructor
    base_poller.InternalPoller.__init__(self)

  #--------------------------------------------------------------------------
  def usage(self):
    opusutil.PrintMsg('F','Interactive Usage: cdb_archive_prep.py [-h] -d delivery_name -p pathname')
    raise UsageError

  #--------------------------------------------------------------------------
  #
  # Name: interactive
  #
  # Description: Performs command line parsing and set-up for
  #              processing in interactive mode.  Runs the
  #              task for the command-line arguments given.
  #
  # Inputs: global sys.argv (command line arguments)
  #          
  #        Will EXIT (status=1) if incorrect command line arguments
  #        are found.  This is a FATAL error.
  #
  # History:
  #   date    opr     who     reason
  # -------- -----  --------  --------------------------------------
  # 09/23/02 xxxxx  MSwam     first version
  #
  #--------------------------------------------------------------------------
  def interactive(self):
    #
    # parse command line arguments
    try:
      opts, args = getopt.getopt(sys.argv[1:], "hd:p:a:1:2:s:", 
                   ["help","delivery_name=","pathname=","archive_dir=",
                    "osf_status="])
    except getopt.GetoptError:
      # print help information and raise exception
      self.usage()

    # initialize
    delivery_name = None
    pathname = None

    # parse options
    for flag, value in opts:
      if flag in ("-h", "--help"):
        self.usage()
      if flag in ("-d", "--delivery_name"):
        delivery_name = value
      if flag in ("-p", "--pathname"):
        self._pathname = value
      if flag in ("-a", "--archive_directory"):
        self._archive_dir = value
      if flag in ("-s", "--osf_status"):
        self._osf_status = value

    # verify required args
    if delivery_name == None or self._pathname == None:
      self.usage()

    # set the directory for file operations to the current directory
    self._inpath = os.getcwd()
    self._trlpath = self._inpath

    # run the computations
    self.do_cdb_archive_prep(delivery_name)

    return

  #-------------------------------------------------------------------------
  # 
  # Name: pipeline_init
  #
  # Description: Initialization for pipeline mode, including reading
  #              resource file parameters for controlling the task.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 09/23/02 xxxxx  MSwam     first version
  # 02/28/05 51433  MSwam     use base_poller
  #
  #-------------------------------------------------------------------------
  def pipeline_init(self):
    # obtain resource items in a dictionary and assign to globals
    #
    try:
      resource_items = base_poller.InternalPoller.get_resource_items(self,
                            ["INPATH",
                             "ARCHIVE_DIR",
                             "OSF_STATUS"
                             ] )
    except:
      opusutil.PrintMsg("E","failed to get required resource entry")
      raise
    #
    opusutil.PrintMsg("I","Resource items read.")
    self._inpath = resource_items['INPATH']
    # trailer is assumed to be in same directory as the catalog
    self._trlpath = self._inpath
    self._archive_dir = resource_items['ARCHIVE_DIR']
    self._osf_status  = resource_items['OSF_STATUS']
   
    # get pathname this task is running in
    self._pathname = self._papp._get_path()

    #
    return

  #-------------------------------------------------------------------------
  # 
  # Name: catalogInserts
  #
  # Description: Returns a list of dataset names read from a CDBS
  #              release catalog.  Only catalog INSERTION (I) entries
  #              are returned (DELETION (D) entries are ignored).
  #
  # Input: delivery_name - name of the delivery to process
  #
  # Output: datasets - list of insertion datasets in a CDBS release catalog
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 09/23/02 xxxxx  MSwam     first version
  # 04/26/05 51433  MSwam     adjust dataset trim to handle synphot as well
  #
  #-------------------------------------------------------------------------
  def catalogInserts(self, delivery_name):
    # form catalog name from delivery name
    catalog_name = self._inpath + os.sep + delivery_name + '.cat'

    # read catalog insert datasets into a list
    datasets = []
    try:
      for line in open(catalog_name, 'r').readlines():
        # split line and save dataset if entry is an insertion
        parts = string.split(line)
        if parts[1] == "I":
          # trim extension from the dataset name (last _ and extra chars)
          pos = parts[0].rfind('_')
          if pos > 0:
            datasets.append(string.lower(parts[0][:pos]))
          else:
            datasets.append(string.lower(parts[0]))
    except IOError:
      opusutil.PrintMsg("E","Failed to read catalog "+catalog_name)
      raise
    #
    return datasets

  #-------------------------------------------------------------------------
  # 
  # Name: setDataId
  #
  # Description: Determines the value for the OSF data id by using
  #              the reference file type (IMAGE or TABLE) from the
  #              first data file found.
  #
  # Input: dataset - name of dataset to process
  #        files - list of files found on disk
  #
  # Output: data_id - value for data_id field of dataset OSF
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 09/23/02 xxxxx  MSwam     first version
  # 06/29/04 51568  MSwam     all WF2 are CDB class
  #
  #-------------------------------------------------------------------------
  def setDataId(self, dataset, files):
    # FITS files
    if len(files) == 0:
      opusutil.PrintMsg("E","No files found for "+dataset)
      raise NoFilesFound

    # read 1st XTENSION keyword from FITS files until found
    filetype = ""
    for aFile in files: 
      opusutil.PrintMsg("D","reading keywords from "+aFile)
      try:
        fd = pyfits.open(aFile)
        filetype = fd[1].header['XTENSION']
        opusutil.PrintMsg("D","found filetype= "+filetype)
      except:
        opusutil.PrintMsg("D","failed to get XTENSION[1] from "+aFile)
        pass
      if filetype != "": break

    # translate keyword value to data id value
    opusutil.PrintMsg("D","using filetype= "+filetype)
    if dataset[-1:] != 'u' and string.find(filetype, "TABLE") != -1:
      return "ctb"
    else:
      # default to IMAGE
      return "cdb"

  #-------------------------------------------------------------------------
  # 
  # Name: do_cdb_archive_prep
  #
  # Description: Prepares a CDBS reference file delivery for archiving
  #              by creating OSFs for each reference file dataset.
  #
  # Input:
  # delivery_name - name of the delivery to process
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 09/23/02 xxxxx  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def do_cdb_archive_prep(self, delivery_name):
    # obtain list of ref file "insert" datasets from catalog file
    datasets = self.catalogInserts(delivery_name)
    opusutil.PrintMsg("D","datasets="+str(datasets))
    
    # loop through datasets
    for dataset in datasets:
       #
       # create a hard link for each file to 
       #  - the archive pick-up area
       #
       filespec = self._inpath + os.sep + dataset + '*.fits'
       files = glob.glob(filespec)
       for aFile in files:
         # link failures are assumed to be duplicates, which are ignored
         # (until proven otherwise)
         try:
           os.link(aFile, self._archive_dir + os.sep + os.path.basename(aFile))
         except OSError:
           opusutil.PrintMsg("W","Failed to create archive_dir link for "+aFile)
           pass

       # set data_id for OSF from file type
       data_id = self.setDataId(dataset, files)

       # create OSF
       try:
         opusutil.PrintMsg("I","creating OSF for dataset="+dataset+
                               " dataid="+data_id+
                               " path="+self._pathname)
         os.system('osf_create -p ' + self._pathname + ' -f ' + dataset + 
                   ' -t ' + data_id + ' -n 000 -s '+self._osf_status)
       except:
         opusutil.PrintMsg('E','OSF creation failed.')
    #
    return

  #-------------------------------------------------------------------------
  # Name: processHaltEvent
  #
  # Description: Handle a HALT event (override base class).
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 02/28/05 51433  MSwam     use base_poller
  #
  #-------------------------------------------------------------------------
  def processHaltEvent(self, event):
    #
    opusutil.PrintMsg("I","Received HALT event. Stopping.")
    raise HaltEvent

  #-------------------------------------------------------------------------
  # Name: processOsfEvent
  #
  # Description: Handle OSF events in pipeline mode.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 04/01/04 49754  MSwam     first version
  # 05/11/04 49754  MSwam     fix success trigger name
  # 08/06/04 49860  Sontag    Changed an IDL field name
  # 09/22/04 51879  MSwam     Use base_poller class
  # 10/29/04 52310  MSwam     remove db closes (only HALT event should)
  # 02/28/05 51433  MSwam     use base_poller
  #
  #-------------------------------------------------------------------------
  def processOsfEvent(self, event):
        # get trigger info
        dataset, data_id = self.getDatasetDataidFromOsfEvent(event)

        try:
          opusutil.OpenTrl(self._trlpath + os.sep + dataset + ".trl")
        except:
          # any exception marks the OSF with OSF_ERROR
          traceback.print_exc()
          #
          # an error occurred during trailer opening, set the OSF to failed
          try:
            opusutil.PrintMsg("E","closing event with OSF_ERROR")
            self._papp.closeEvent("OSF_ERROR", event.eid)
          except:
            traceback.print_exc()
            opusutil.PrintMsg("E","Failed to close OSF_ERROR event")
          return

        opusutil.PrintMsg("I","----- CDBARC started for: "+dataset+" -----")
        try:
          mode = self.do_cdb_archive_prep(dataset)
        except:
          # any other exception marks the OSF with OSF_ERROR
          traceback.print_exc()
          #
          # an error occurred during processing, set the OSF
          try:
            opusutil.PrintMsg("E","closing event with OSF_ERROR")
            self._papp.closeEvent("OSF_ERROR", event.eid)
          except:
            traceback.print_exc()
            opusutil.PrintMsg("E","Failed to close OSF_ERROR event")
          opusutil.PrintMsg("I","----- CDBARC ended for: "+dataset+" -----")
          opusutil.CloseTrl()
          return

        try:
          self._papp.closeEvent("OSF_NORMAL", event.eid)
        except:
            traceback.print_exc()
            opusutil.PrintMsg("E","Failed to close OSF_NORMAL event")
        opusutil.PrintMsg("I","----- CDBARC ended for: "+dataset+" -----")
        opusutil.CloseTrl()

#-------------------------------------------------------------------------
# The main routine. Handles interactive and pipeline mode.
#
# History:
# -------- -----  --------  --------------------------------------
# 02/28/05 51433  MSwam     use base_poller
#
#-------------------------------------------------------------------------
#
# initialize
try:
  poller = ArchivePrep()
  poller.initialize()
  poller.pipeline_init()
#
except base_poller.InteractiveMode:
  opusutil.PrintMsg("I","Running command-line mode.")
  try:
    poller.interactive()
  except UsageError:
    sys.exit(NOT_OK)
  except:
    traceback.print_exc()
    opusutil.PrintMsg("F","Command-line mode Failed.")
    sys.exit(NOT_OK)
  sys.exit(OK)
#
except:
  traceback.print_exc()
  opusutil.PrintMsg("E","initialization failed")
  sys.exit(NOT_OK)

opusutil.PrintMsg("I","Pipeline set-up complete.")
try:
  poller.run()
except HaltEvent:
  sys.exit(OK)
except:
  traceback.print_exc()
  opusutil.PrintMsg("E","pipeline_process failed")
  sys.exit(NOT_OK)
